/*
///////////////////////////////////////////////////////////////
//                                                           //
//  pitry.c   --      To calculate value of Pi squared       //
//  ver 1.0                                                  //
//                                                           //
//  Application:    CPS 615, Assignment #3                   //
//  Author:         Sunil Samuel                             //
//                  slsamuel@mailbox.syr.edu                 //
//                                                           //
///////////////////////////////////////////////////////////////
*/

/* Include all header files required */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

/* Initializing the number of processors in the grid to
   be 4, of which 2 are along the x-axis and 2 are along the 
   y-axis. */

#define xprocessors 2
#define yprocessors 2

/* Start of the program */
double sqrt(double);

main(int argc, char* argv[]) {
	int size, reorder;
	int dims[2], periods[2], coords[2];
	int rank, k, n;
        double pisquared, h, h1, h2, pi2, x, y, f1, f2, f3, f4;
	double total[10]; /* this stores the values of 
               the pi squares calculated by the individual processors.*/
	double i, j;
	double xbegin;             /* initial coordinate of x   */
	double xend;               /* final coordinate of x     */
	double ybegin;             /* initial coordinate of y   */
	double yend;               /* final coordinate of y     */
	 /* the value of pi square that each processor calculates                */
	 /* The final value of pi square after all the values are added together */
 

    MPI_Comm grid_comm;

    /* This function starts up MPI, it must be called before any other MPI function is called */
    MPI_Init(&argc, &argv);

    /* Returns the no of processors in the underlying comm */
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if (size!=xprocessors*yprocessors) {
	printf("The  number of processors don't match. It has to be 4");
	exit(1);
    }

    periods[0] = 1;
    periods[1] = 1;

    dims[0] = yprocessors;
    dims[1] = xprocessors;

    reorder = 0;
    n = 10;
 
 /* creates a new communicator with cartesian coordinate structure on 
    the processes*/  
    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder, &grid_comm);

 /* Returns the rank of the calling process in the group underlying the
    communicator comm */
    MPI_Comm_rank(grid_comm, &rank);

 /* Returns the coordinates of the process with rank rank in the
    cartesian communicator comm */ 
    MPI_Cart_coords(grid_comm, rank, 2, coords);
    

    h1 = 1.0/((double)xprocessors*(double)n);
    h2 = 1.0/((double)yprocessors*(double)n);
    h = (h1+h2)/2.0;
	/* Calculates begining coordinate of x*/
    xbegin = (double)coords[0]*(double)n; 
	/* Calculates begining coordinate of y*/
    ybegin = (double)coords[1]*(double)n; 
	/* Calculates ending coordinate of x*/
    xend = xbegin + (double)n;  
	/* Calculates ending coordinate of y*/
    yend = ybegin + (double)n; 
    pisquared = 0.0;

    /* Loop over all the intervals along the x and y axis. 
       i is used to traverse along the x-axis and j to traverse.*/
 
    for (i=xbegin; i<xend; i++) {
	for (j=ybegin; j<yend; j++) {

	        x = i * h1;
		y = j * h2;
		f1 = 1.0/(1.0+x*x);
		f2 = 1.0/(1.0+y*y);
		f3 = 1.0/(1.0+(y+h2)*(y+h2));
		f4 = 1.0/(1.0+(x+h1)*(x+h1));
		pisquared+=(h1*h2)*(f1*f2+f1*f3+f2*f4+f3*f4)/4.0;
	}
    }
 	  

   pisquared = 16.0 * pisquared;
   
   MPI_Gather(&pisquared,1,MPI_DOUBLE,&total,1,MPI_DOUBLE,0,grid_comm);

   pi2 = 0.0; /* Initialize pi2 to zero. pi2 will contain the final
                 value of pi squared */

   /* Rank 0 is taken to the master processor and this processor
      adds the values of pi that have been calculated by the
      individual processors. The final value of pi square is obtained
      when all the values are added. */

   if(rank == 0) { 
      for(k = 0; k < size; k++)
        pi2 += total[k];
      printf("The  Final value of pi squared is %.16f  and pi= %0.16f\n", pi2,
	     sqrt(pi2));
   }

   /* this function shuts down MPI. No MPI function may be 
      called after this function */
   MPI_Finalize();
}

